@@ -48,6 +48,7 @@ extern gchar *GdmUserAuthFB;
 extern gint  GdmUserMaxFile;
 extern gint  GdmRelaxPerms;
 extern gboolean GdmDebug;
+extern gboolean GdmNeverPlaceCookiesOnNFS;
 
 static void
 display_add_error (GdmDisplay *d)
@@ -215,7 +216,7 @@ gdm_auth_secure_display (GdmDisplay *d)
 
 	    if G_UNLIKELY (af_gdm == NULL) {
 		    gdm_error (_("%s: Cannot safely open %s"),
-			       "gdm_auth_user_remove",
+			       "gdm_auth_secure_display",
 			       d->authfile_gdm);
 
 		    g_free (d->authfile_gdm);
@@ -232,7 +233,7 @@ gdm_auth_secure_display (GdmDisplay *d)
 
 	    if G_UNLIKELY (af == NULL) {
 		    gdm_error (_("%s: Cannot safely open %s"),
-			       "gdm_auth_user_remove",
+			       "gdm_auth_secure_display",
 			       d->authfile);
 
 		    g_free (d->authfile);
@@ -433,7 +434,6 @@ get_local_auth_error:
     return NULL;
 }
 
-
 static gboolean
 try_open_append (const char *file)
 {
@@ -493,6 +493,7 @@ gdm_auth_user_add (GdmDisplay *d, uid_t user, const char *homedir)
     gboolean automatic_tmp_dir = FALSE;
     gboolean authdir_is_tmp_dir = FALSE;
     gboolean locked;
+    gboolean user_auth_exists;
 
     if (!d)
 	return FALSE;
@@ -536,17 +537,38 @@ try_user_add_again:
     else
 	    d->userauth = g_build_filename (authdir, GdmUserAuthFile, NULL);
 
+    user_auth_exists = (d->userauth != NULL &&
+			access (d->userauth, F_OK) == 0);
+
     /* Find out if the Xauthority file passes the paranoia check */
+    /* Note that this is not very efficient, we stat the files over
+       and over, but we don't care, we don't do this too often */
     if (automatic_tmp_dir ||
 	authdir == NULL ||
+
+	/* first the standard paranoia check (this checks the home dir
+	 * too which is useful here) */
 	! gdm_file_check ("gdm_auth_user_add", user, authdir, GdmUserAuthFile, 
 			  TRUE, FALSE, GdmUserMaxFile, GdmRelaxPerms) ||
+
+	/* now the auth file checking routine */
+	! gdm_auth_file_check ("gdm_auth_user_add", user, d->userauth, TRUE /* absentok */, NULL) ||
+
+	/* now see if we can actually append this file */
 	! try_open_append (d->userauth) ||
 
 	/* try opening as root, if we can't open as root,
 	   then this is a NFS mounted directory with root squashing,
 	   and we don't want to write cookies over NFS */
-	! try_open_read_as_root (d->userauth)) {
+	(GdmNeverPlaceCookiesOnNFS && ! try_open_read_as_root (d->userauth))) {
+
+        /* if the userauth file didn't exist and we were looking at it,
+	   it likely exists now but empty, so just whack it
+	   (it may not exist if the file didn't exist and the directory
+	    was of wrong permissions, but more likely this is
+	    file on NFS dir with root-squashing enabled) */
+        if ( ! user_auth_exists && d->userauth != NULL)
+		unlink (d->userauth);
 
 	/* No go. Let's create a fallback file in GdmUserAuthFB (/tmp)
 	 * or perhaps GdmUserAuth directory (usually would be /tmp) */
@@ -581,6 +603,8 @@ try_user_add_again:
 	    return FALSE;
 	}
 
+	d->last_auth_touch = time (NULL);
+
 	af = fdopen (authfd, "w");
     }
     else { /* User's Xauthority file is ok */
@@ -701,6 +725,7 @@ gdm_auth_user_remove (GdmDisplay *d, uid_t user)
     FILE *af;
     gchar *authfile;
     gchar *authdir;
+    mode_t oldmode;
 
     if G_UNLIKELY (!d || !d->userauth)
 	return;
@@ -726,12 +751,21 @@ gdm_auth_user_remove (GdmDisplay *d, uid_t user)
     authfile = g_path_get_basename (d->userauth);
     authdir = g_path_get_dirname (d->userauth);
 
+    if (ve_string_empty (authfile) ||
+	ve_string_empty (authdir)) {
+	    g_free (authdir);
+	    g_free (authfile);
+	    return;
+    }
+
     /* Now, the cookie file could be owned by a malicious user who
      * decided to concatenate something like his entire MP3 collection
      * to it. So we better play it safe... */
 
     if G_UNLIKELY ( ! gdm_file_check ("gdm_auth_user_remove", user, authdir, authfile, 
-			   TRUE, FALSE, GdmUserMaxFile, GdmRelaxPerms)) {
+			   TRUE, FALSE, GdmUserMaxFile, GdmRelaxPerms) ||
+		    /* be even paranoider with permissions */
+		    ! gdm_auth_file_check ("gdm_auth_user_remove", user, d->userauth, FALSE /* absentok */, NULL)) {
 	    g_free (authdir);
 	    g_free (authfile);
 	    gdm_error (_("%s: Ignoring suspiciously looking cookie file %s"),
@@ -751,7 +785,9 @@ gdm_auth_user_remove (GdmDisplay *d, uid_t user)
 	return;
     }
 
+    oldmode = umask (077);
     af = gdm_safe_fopen_ap (d->userauth);
+    umask (oldmode);
 
     if G_UNLIKELY (af == NULL) {
 	XauUnlockAuth (d->userauth);
@@ -825,6 +861,8 @@ gdm_auth_purge (GdmDisplay *d, FILE *af, gboolean remove_when_empty)
 {
     Xauth *xa;
     GSList *keep = NULL, *li;
+    mode_t oldmode;
+    int cnt;
 
     if G_UNLIKELY (!d || !af)
 	return af;
@@ -837,6 +875,8 @@ gdm_auth_purge (GdmDisplay *d, FILE *af, gboolean remove_when_empty)
      * temporary file issues. Then remove any instance of this display
      * in the cookie jar... */
 
+    cnt = 0;
+
     while ( (xa = XauReadAuth (af)) != NULL ) {
 	GSList *li;
 	/* We look at the current auths, but those may
@@ -853,6 +893,11 @@ gdm_auth_purge (GdmDisplay *d, FILE *af, gboolean remove_when_empty)
 	}
 	if (xa != NULL)
 		keep = g_slist_append (keep, xa);
+
+	/* just being ultra anal */
+	cnt ++;
+	if (cnt > 500)
+		break;
     }
 
     fclose (af);
@@ -863,7 +908,9 @@ gdm_auth_purge (GdmDisplay *d, FILE *af, gboolean remove_when_empty)
 	    return NULL;
     }
 
+    oldmode = umask (077);
     af = gdm_safe_fopen_w (d->userauth);
+    umask (oldmode);
 
     /* Write out remaining entries */
     for (li = keep; li != NULL; li = li->next) {
